/****************************************************  Filter Functions  *****************************************************

	Red Rapids
	797 North Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are used to configure a variety of filters available on SigStream products.

	Rev History:  01-10-2017:	Inital release.

	Copyright 2017, Red Rapids, Inc.

****************************************************  Filter Functions  *****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "RRadapter0.h"
#include "sigstream.h"
#include "dma_functions.h"
#include "common_functions.h"
#include "channel_functions.h"
#include "filter_functions.h"


// Define the unique characteristics of every filter type available in the product family.
size_t Filter[FilterTypes][6] =
{
	// Type 0 FIR filter definition.
	{
		87,		// Number of Taps			Size of the filter in number of taps or coefficients.
		0,		// Reload Order				Unique code to convey the coefficient reload order of the filter.
		24,		// Coeffiecient Size		Size of each coefficient in bits.
		1,		// Symmetry					Code to convey symmetry properties of the filer (0 = asymmetric, 1 = symmetric, 2 = antisymmetric).
		1,		// Minimum Resample			Minimum downsample ratio (input:output) or upsample ratio (output:input).
		0		// Reserved
	},
	// Type 1 FIR filter definition.
	{
		127,	// Number of Taps			Size of the filter in number of taps or coefficients.
		1,		// Reload Order				Unique code to convey the coefficient reload order of the filter.
		24,		// Coeffiecient Size		Size of each coefficient in bits.
		1,		// Symmetry					Code to convey symmetry properties of the filer (0 = asymmetric, 1 = symmetric, 2 = antisymmetric).
		1,		// Minimum Resample			Minimum downsample ratio (input:output) or upsample ratio (output:input).
		0		// Reserved
	},
	// Type 2 FIR filter definition.
{
		255,	// Number of Taps			Size of the filter in number of taps or coefficients.
		2,		// Reload Order				Unique code to convey the coefficient reload order of the filter.
		23,		// Coeffiecient Size		Size of each coefficient in bits.
		1,		// Symmetry					Code to convey symmetry properties of the filer (0 = asymmetric, 1 = symmetric, 2 = antisymmetric).
		1,		// Minimum Resample			Minimum downsample ratio (input:output) or upsample ratio (output:input).
		0		// Reserved
}
};

// Load coefficients from a file into the designated filter.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long FilterConfigure(s_Adapter *p_HW, unsigned long ChannelNumber, unsigned long long Select, char *p_File)
{
	unsigned long Status = 0;
	unsigned long long *p_coefficients;
	size_t log2gain = 0;
	size_t resample = 0;
	size_t type = 0;

	// Get the type of the selected filter.
	type = GetFilterType(p_HW, ChannelNumber, Select);

	// Allocate storage for all of the coefficients.
	p_coefficients = (unsigned long long*)malloc(Filter[type][0] * sizeof(unsigned long long));

	// Parse the configuration file for filter characteristics.
	Status += FilterParser(type, p_coefficients, &log2gain, &resample, p_File);

	// Configure the filter if there were no prior errors.
	if (!Status)
	{
		// Load the filter coefficients.
		Status += LoadFilterCoefficients(p_HW, ChannelNumber, Select, type, p_coefficients);

		// Load the filter gain setting.
		Status += LoadFilterGain(p_HW, ChannelNumber, Select, type, log2gain);

		// Load the filter resampler setting.
		Status += LoadFilterResampler(p_HW, ChannelNumber, Select, type, resample);
	}

	// Free the coefficient memory if it was successfully allocated.
	if (p_coefficients != NULL)
		free(p_coefficients);

	return(Status);
}

// Create a coefficient file name of the form filter<#1>_type<#2>.txt, where #1 is the number and #2 is the type.
void FilterFilename(s_Adapter *p_HW, unsigned long ChannelNumber, unsigned long long Select, char *p_File)
{
	size_t type = 0;
	char number[3];

	// Get the type of the selected FIR.
	type = GetFilterType(p_HW, ChannelNumber, Select);

	// Create a coefficient ch0 of the form filter<#1>_type<#2>.txt, where #1 is the number and #2 is the type.
	// Note: Assumes files are located in directory ..\coefficients (Windows) or ../coefficents (Linux).
	strcpy(p_File, "..");
	strcat(p_File, SEPARATOR);
	strcat(p_File, "filters");
	strcat(p_File, SEPARATOR);
	strcat(p_File, "filter");
	sprintf(number, "%llu", Select);
	strcat(p_File, number);
	strcat(p_File, "_type");
	sprintf(number, "%zu", type);
	strcat(p_File, number);
	strcat(p_File, ".txt");
}

// Search for a keyword in the configuration file.
unsigned long FilterKeyword(char *Keyword)
{
	unsigned long keyword_id = 0;

	// Create a coefficient ch0 of the form fir<#1>_type<#2>.txt, where #1 is the number and #2 is the type.
	if (!strncmp(Keyword, "Coef", 4) || !strncmp(Keyword, "coef", 4) || !strncmp(Keyword, "COEF", 4)) keyword_id = 1;
	if (!strncmp(Keyword, "Log2", 4) || !strncmp(Keyword, "log2", 4) || !strncmp(Keyword, "LOG2", 4)) keyword_id = 2;
	if (!strncmp(Keyword, "Resa", 4) || !strncmp(Keyword, "resa", 4) || !strncmp(Keyword, "RESA", 4)) keyword_id = 3;
	return(keyword_id);
}

// Search for a keyword in the configuration file.
unsigned long FilterParser(size_t Type, unsigned long long *p_Coefficients, size_t *p_Log2Gain, size_t *p_Resample, char *p_File)
{
	FILE *fp;
	unsigned long Status = 0;
	size_t cnt = 0;
	char inputline[16];

	// Process the configuration file only if it opens successfully.
	if ((fp = fopen(p_File, "r")) != NULL)
	{
		// Parse the configuration file for keywords.
		while (fgets(inputline, 16, fp) != NULL)
			// Process the keywords.
			switch (FilterKeyword(inputline))
			{

				// Read the coefficients.
			case 1:
			{
				// Load the coefficients only if memory allocation succeeds.
				if (p_Coefficients != NULL)
				{
					// Read and store all of the available coefficients.
					for (cnt = 0; cnt < Filter[Type][0]; cnt++)
						if (fscanf(fp, "%llu", &p_Coefficients[cnt]) == EOF) Status++;
				}
				else
					Status++;
				break;
			}

			// Read the base 2 logarithm of the gain.
			case 2: {if (fscanf(fp, "%zu", p_Log2Gain) == EOF) Status++; break; }

					// Read the resample factor.
			case 3: {if (fscanf(fp, "%zu", p_Resample) == EOF) Status++; break; }

					// Do nothing if not a recognized keyword.
			default: {break; }
			}
		// Close the file.
		fclose(fp);
	}
	else
		Status++;

	return(Status);
}

// Report the number of filters assigned to the requested channel.
// Return the 8-bit filter count.
size_t GetFilterCount(s_Adapter *p_HW, unsigned long ChannelNumber)
{
	unsigned long long Value = 0;

	// Read the filter status.
	Read(p_HW, BAR2, ChannelNumber << 12 | FILTER_CONFIG_STAT, &Value);

	return((size_t)((Value >> 32) & 0x00000000000000FFull));
}

// Report the unique type ID associated with the selected filter.
// Return the 8-bit type ID.
size_t GetFilterType(s_Adapter *p_HW, unsigned long ChannelNumber, unsigned long long Select)
{
	unsigned long long Value = 0;

	// Select the specific to identify.
	Write(p_HW, BAR2, ChannelNumber << 12 | FILTER_CONFIG_CMD, (Select - 1) << 4);
	// Read the filter ID.
	Read(p_HW, BAR2, ChannelNumber << 12 | FILTER_CONFIG_STAT, &Value);

	return((size_t)((Value >> 10) & 0x00000000000000FFull));
}

// Load the filter coefficients.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long LoadFilterCoefficients(s_Adapter *p_HW, unsigned long ChannelNumber, unsigned long long Select, size_t Type, unsigned long long *p_Coefficients)
{
	unsigned long Status = 0;
	unsigned long long Value = 0;
	size_t index = 0;
	size_t symmetry = 2;

	// Determine if coefficients are asymmetric (0).
	for (index = 0; index < Filter[Type][0]/2; index++)
		if ((p_Coefficients[index] != p_Coefficients[Filter[Type][0]-1-index]) && (p_Coefficients[index] != -1* p_Coefficients[Filter[Type][0]-1-index]))
			symmetry = 0;
	// If the coefficients are not asymmetric, determine wheter they are symmetric (1) or antisymmetric (2).
	if ((symmetry != 0) && (p_Coefficients[0] == p_Coefficients[Filter[Type][0]-1]))
		symmetry = 1;

	// Make sure the supplied coefficients match the characteristics.
	if (symmetry == Filter[Type][3])
		// Reorder the coefficents and reload the FIR.
		for (index = 0; index <= Filter[Type][0]/2; index++)
		{
			// Read the status register to determine whether there is a command pending, which should be impossible.
			Read(p_HW, BAR2, ChannelNumber << 12 | FILTER_CONFIG_STAT, &Value);
			// Load the next coefficient if there is no command pending.
			if (!(Value & 0x0000000000000100ull) && !Status)
			{
				// Select the next coefficient based on a specific reload order.
				switch (Filter[Type][1])
				{
					// Forward sequential.
					case 0: { Value = p_Coefficients[index]; break; }
					// Reverse modulo 2.
					case 1: { Value = p_Coefficients[((Filter[Type][0] + 1) / 2) - 2 - index + (2 * (index % 2))]; break; }
					// Reverse modulo 4.
					case 2: { Value = p_Coefficients[((Filter[Type][0] + 1) / 2) - 4 - index + (2 * (index % 4))]; break; }
					// Flag an error if there is no match.
					default: { Value = 0; Status++; break; }
				}
				// Write the coefficient reload value.
				if (index == Filter[Type][0]/2)
					// Enable the tlast bit.
					Write(p_HW, BAR2, ChannelNumber << 12 | FILTER_CONFIG_DATA, 0x0000000001000000ull | Value);
				else
					// Do not enable the tlast bit.
					Write(p_HW, BAR2, ChannelNumber << 12 | FILTER_CONFIG_DATA, Value);
				// Write the command to execute the reload operation of the FIR.
				Write(p_HW, BAR2, ChannelNumber << 12 | FILTER_CONFIG_CMD, ((Select-1) << 4) | 0x0000000000000001ull);
			}
			// There is a command pending which shouldn't be possible so return an error.
			else
				Status++;
		}
	// Return an error if the coefficents don't match the expected symmetry.
	else
		Status++;

	// Write a zero config value since the only supports one coefficent set.
	Write(p_HW, BAR2, ChannelNumber << 12 | FILTER_CONFIG_DATA, 0x0000000000000000ull);
	// Activate the latest coefficient reload values, even if the reload was not successfull.
	Write(p_HW, BAR2, ChannelNumber << 12 | FILTER_CONFIG_CMD, ((Select-1) << 4) | 0x0000000000000003ull);

	return(Status);
}

// Load the filter gain.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long LoadFilterGain(s_Adapter *p_HW, unsigned long ChannelNumber, unsigned long long Select, size_t Type, size_t Log2Gain)
{
	unsigned long Status = 0;
	unsigned long long bit_shift = 0;

	// Set the bit shift equal to the maximum number of coefficent magnitude bits, not including the sign bit.
	bit_shift = Filter[Type][2] - 1;

	// The bit shifter has a range of zero to seven bits from full scale.
	if (Log2Gain < Filter[Type][2]-1)
	{
		bit_shift = 0;
		Status++;
	}
	else if (Log2Gain > Filter[Type][2]+8)
	{
		bit_shift = 7;
		Status++;
	}
	else
		// Select the significant bits from the 24-bit output that will result in maximum dynamic range.
		bit_shift = Log2Gain - bit_shift;

	// Set the bit shifter value to output the most significant MSBs.
	Write(p_HW, BAR2, ChannelNumber << 12 | FILTER_CONFIG_DATA, bit_shift);

	// Load the gain value.
	Write(p_HW, BAR2, ChannelNumber << 12 | FILTER_CONFIG_CMD, ((Select-1) << 4) | 0x0000000000000005ull);

	return(Status);
}

// Load the filter resmapler.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long LoadFilterResampler(s_Adapter *p_HW, unsigned long ChannelNumber, unsigned long long Select, size_t Type, size_t Resample)
{
	unsigned long Status = 0;

	// Set the resampler to the requested value.
	if (Resample == 0)
		// Bypass the entire filter stage.
		Write(p_HW, BAR2, ChannelNumber << 12 | FILTER_CONFIG_DATA, 0x0000000000010000ull);
	// Determine whether the requested resampler value is at least the minimum required.
	else if (Resample < Filter[Type][4])
		{
			// Set the resampler to zero, which will result in the built-in minimum value.
			Write(p_HW, BAR2, ChannelNumber << 12 | FILTER_CONFIG_DATA, 0x0ull);
			Status++;
		}
	else
		{
			// Set the resampler to a zero based value after removing the built-in minimum value.
			Write(p_HW, BAR2, ChannelNumber << 12 | FILTER_CONFIG_DATA, (Resample/Filter[Type][4])-1);
			// Set an error if the requested resampler value is not a multiple of the minimum required.
			if (Resample%Filter[Type][4] != 0)
				Status++;
		}

	// Load the resampler value.
	Write(p_HW, BAR2, ChannelNumber << 12 | FILTER_CONFIG_CMD, ((Select-1) << 4) | 0x0000000000000007ull);

	return(Status);
}
