/****************************************************  Utility Functions  *****************************************************

	Red Rapids
	797 North Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are specific to the Signal Stream application example.

	Rev History:  01-10-2017:	Modified DataSave() to process 32-bit data items and FIFOResidue() to correctly report 128-bit FIFO residue. Removed the error check from ProcessChannels() if the DMA engine is disabled.
				  08-21-2014:	Inital release.

	Copyright 2014, Red Rapids, Inc.

****************************************************  Utility Functions  *****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "RRadapter0.h"
#include "sigstream.h"
#include "dma_functions.h"
#include "common_functions.h"
#include "channel_functions.h"

// Report the current state of the synchronizer.
// 0: Idle
// 1: Initialize/Prime
// 2: Start Trigger
// 4: Start Delay
// 8: Synch/Stop Trigger
// 16: Stop Delay
// 32: Purge/Residue
// 64: Cycle
// 128: Scheduler/Stop Bench
// 255: Undefined
unsigned long ChannelState(s_Adapter *p_HW, unsigned long ChannelNumber)
{
	unsigned long long Value = 0;
	
	Read(p_HW, BAR2, ChannelNumber << 12 | CHANNEL_STATE, &Value);
	
	return((unsigned long)(Value & 0x00000000000000FFull));
}

// Clear the PCI latency performance measurement.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ClearLatency(s_Adapter *p_HW)
{
	unsigned long Status = 0;

	// Clear the PCI latency performance register.
	Status += Write(p_HW, BAR0, PCI_LATENCY, 0x0ull);

	return(Status);
}

// Preload a transmitter buffer with sine wave data samples.
// Return of zero indicates that the buffer contains an integer number of sine wave periods, non-zero indicates a fractional number of periods.
unsigned long DataLoad(void **p_PageAddresses, s_Channel TXChannel, double FsDivisor, size_t DataItemSize)
{
	unsigned long long max = 0xFFFFFFFFFFFFFFFFull;
	size_t page_count = 0;
	size_t record_count = 0;
	size_t records_per_page = 0;
	size_t data_items_per_record = 0;
	size_t data_items_count = 0;
	unsigned long long *virtual_address;
	double phase = 0;
	double phase_select = 0;
	double data_item_size_dbl = 0;

	// Cast the DataItemSize variable for use with other data types.
	data_item_size_dbl = (double)DataItemSize; 

	// Compute the number of 64-bit records per DMA page.
	records_per_page = TXChannel.DMABuffer.BurstSizeB * TXChannel.DMABuffer.BurstCount / 8;

	// Compute the number of data items per 64-bit record.
	data_items_per_record = (size_t)floor(64 / data_item_size_dbl);

	// Sequence through the number of DMA pages per mark since this one chapter will be used for every DMA buffer refresh.
	for(page_count=0; page_count < TXChannel.DMABuffer.PagesPerMark; page_count++)
	{
		// Get the virtual address of the first 64-bit record in the current page.
		virtual_address = (unsigned long long *)p_PageAddresses[page_count];

		// Sequence through every 64-bit record in the page.
		for(record_count = 0; record_count < records_per_page; record_count++)
		{
			// Clear the current 64-bit record.
			*(virtual_address + record_count) = 0x0ull;
			
			for(data_items_count = 1; data_items_count <= data_items_per_record; data_items_count++)
			{
				// Compute the amplitude of the DAC sample based on phase and write it as the current data item to the payload record.
				*(virtual_address + record_count) |= (((max >> (64-DataItemSize)) & (unsigned long long)((pow(2,data_item_size_dbl-1)-1)*sin(2*3.141592654*phase/FsDivisor))) << (64-(data_items_count*DataItemSize)));

				// Compute the phase of the next sample.
				phase = fmod(++phase_select,FsDivisor);
				// Restart the phase select counter if the phase is zero.
				if (!phase) phase_select = 0;
			}
		}
	}
	if (phase_select)
		return 1;
	else
		return 0;
}

// Write the contents of the requested number of DMA pages associated with the requested channel number to the specified file.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long DataSave(void **p_PageAddresses, s_Channel RXChannel, s_TODSettings TODClock, char *p_File, size_t PageCount, size_t PacketInfo)
{
	FILE *fp;
	size_t pagecnt = 0;
	size_t payload_index = 32;
	size_t packet_counter = 0;
	size_t index = 0;
	size_t record_count = 0;
	size_t field_count = 0;
	size_t current_field = 0;
	size_t records_per_page = 0;
	size_t next_field[8] = {0,0,0,0,0,0,0,0};
	unsigned long long *virtual_address;
	unsigned long field;
	signed long sample = 0;
	unsigned long residue = 0;

	// Open the specified data file.
	fp = fopen(p_File, "w");

	// Data will be packetized if Packet Size is not zero.
	if (RXChannel.IFdpktFormat.PacketSize)
	{
		// Create a linked list of expected packet record types based on the packet header options selected.
		if (RXChannel.IFdpktFormat.StreamIDSelect) {next_field[index] = 1; index = 1;}
		if (RXChannel.IFdpktFormat.ClassIDSelect) {next_field[index] = 2; index = 3;}
		if (RXChannel.IFdpktFormat.TSISelect) {next_field[index] = 4; index = 4;}
		if (RXChannel.IFdpktFormat.TSFSelect) {next_field[index] = 5; index = 6;}
		next_field[index] = 7;
	}
	// Data is not packetized, begin processing samples immediately.
	else
		current_field = 7;

	// Process records from the buffer only if the file open was successful.
	if(fp != NULL)
	{
		// Compute the number of 64-bit records per DMA page.
		records_per_page = RXChannel.DMABuffer.BurstSizeB * RXChannel.DMABuffer.BurstCount / 8;

		// Sequence through the DMA pages.
		for(pagecnt=0; pagecnt < PageCount; pagecnt++)
		{
			// Get the virtual address of the first 64-bit record in the current page.
			virtual_address = (unsigned long long *)p_PageAddresses[pagecnt];

			// Sequence through every 64-bit record in the page.
			for(record_count = 0; record_count < records_per_page; record_count++)
				// There are two 32-bit VITA 49 fields per 64-bit record.
				for(field_count = 0; field_count < 2; field_count++)
				{
					// Select the upper or lower 32 bits of the 64-bit record for the current field.
					if (field_count)
						// Bits 63:32 contain the next field.
						field = (unsigned long)(*(virtual_address + record_count) >> 32);
					else
						// Bits 31:0 contain the next field.
						field = (unsigned long)*(virtual_address + record_count);

					// Process the current record based on the type of information it conveys.
					switch (current_field)
					{
						// Process the Header field of a packet. (The current record must be a valid header or it will be discarded until a valid header is found.)
						case 0:
						{
							// Note: This bit pattern is not unique to a header, it is possible (but unlikely) for a data sequence to produce an unexpected match.
							if	(((field & 0xFFFF) == RXChannel.IFdpktFormat.PacketSize) &&		// Check for a match to the Packet Size field.
							((RXChannel.IFdpktFormat.TSFSelect == 1 && ((field >> 20) & 0x3) == TODClock.TSFCode) ||	// Check for a match to the TSF field.
								(RXChannel.IFdpktFormat.TSFSelect == 0 && ((field >> 20) & 0x3) == 0)) &&				// Check for a TSF field set to zero.
							((RXChannel.IFdpktFormat.TSISelect == 1 && ((field >> 22) & 0x3) == TODClock.TSICode) ||	// Check for a match to the TSI field.
								(RXChannel.IFdpktFormat.TSISelect == 0 && ((field >> 22) & 0x3) == 0)) &&				// Check for a TSI field set to zero.
							(((field >> 24) & 0x3) == 0) &&										// Check for zeros in the reserved field.
							(((field >> 26) & 0x1) == RXChannel.IFdpktFormat.TrailerSelect) &&	// Check for a match to the Trailer field.
							(((field >> 27) & 0x1) == RXChannel.IFdpktFormat.ClassIDSelect) &&	// Check for a match to the Class ID field.
							(((field >> 28) & 0x1) == RXChannel.IFdpktFormat.StreamIDSelect) &&	// Check for a match to the Stream ID field.
							(((field >> 29) & 0x0) == 0))										// Check for a match to the Packet Type field.
							{
								// Print header information to the file.
								if (PacketInfo) fprintf(fp, "Packet Number:  %lu\n", (field >> 16) & 0xF);
								packet_counter = 1;
								current_field = next_field[0];
								residue = 0;
								payload_index = 32;
							}
							break;
						}

						// Process the optional Stream Identifier field of a packet.
						case 1: {if (PacketInfo) fprintf(fp, "Stream Identifier:  x%08lX\n", field); packet_counter++; current_field = next_field[1]; break;}

						// Process the first half of the optional Class Identifier field of a packet.
						case 2: {if (PacketInfo) fprintf(fp, "Class Identifier:  x%08lX", field); packet_counter++; current_field = 3; break;}

						// Process the second half of the optional Class Identifier field of a packet.
						case 3: {if (PacketInfo) fprintf(fp, "%08lX\n", field); packet_counter++; current_field = next_field[3]; break;}

						// Process the optional Integer-seconds Timestamp field of a packet.
						case 4: {if (PacketInfo) fprintf(fp, "Integer-seconds Timestamp:  x%08lX\n", field); packet_counter++; current_field = next_field[4]; break;}

						// Process the first half of the optional Fractional-seconds Timestamp field of a packet.
						case 5: {if (PacketInfo) fprintf(fp, "Fractional-seconds Timestamp:  x%08lX", field); packet_counter++; current_field = 6; break;}

						// Process the second half of the optional Fractional-seconds Timestamp field of a packet.
						case 6: {if (PacketInfo) fprintf(fp, "%08lX\n", field); packet_counter++; current_field = next_field[6]; break;}

						// Process the Data Payload field of a packet or a raw stream.
						case 7:
						{
							// Sequence through every sample in the 32-bit payload.
							while (payload_index >= RXChannel.PayloadFormat.ItemPackingFieldSize)
							{
								// Decrement the payload index to extract the next sample in the 32-bit payload.
								payload_index -= (size_t)RXChannel.PayloadFormat.ItemPackingFieldSize;
								// Extract the next sample value from the 32-bit payload and append any packing residue that may be present from the last payload.
								sample = (signed long)(((field >> payload_index) & (0xFFFFFFFF >> (32 - RXChannel.PayloadFormat.ItemPackingFieldSize))) | residue);
								// Clear the residue for the remainder of the current payload.
								residue = 0;
								// Sign extend negative numbers.
								if (sample & (1 << (RXChannel.PayloadFormat.ItemPackingFieldSize - 1))) sample |= (-1 << (RXChannel.PayloadFormat.ItemPackingFieldSize - 1));
								// Print the sample value to the file.
								fprintf(fp, "%ld\n", sample);
							}

							// Extract the residue and adjust the payload index if packing is enabled and there is residue present in the current payload.
							if (RXChannel.PayloadFormat.PayloadPacking && payload_index)
							{
								residue = (field & (0xFFFFFFFF >> (32 - payload_index))) << (RXChannel.PayloadFormat.ItemPackingFieldSize - payload_index);
								payload_index += 32;
							}
							// Clear the residue and reset the payload index if packing is not enabled or there is no residue present in the current payload.
							else
							{
								residue = 0;
								payload_index = 32;
							}

							// Track the number of payloads processed if packets are enabled (Packet Size != 0).
							if (RXChannel.IFdpktFormat.PacketSize)
							{
								packet_counter++;
								// Determine whether this is the last payload in a packet when the optional trailer is selected.
								if (RXChannel.IFdpktFormat.TrailerSelect && (packet_counter == (RXChannel.IFdpktFormat.PacketSize - 1)))
									current_field = 8;
								// Determine whether this is the last payload in a packet when the optional trailer is not selected.
								else if (packet_counter == RXChannel.IFdpktFormat.PacketSize)
									current_field = 0;
							}
							break;
						}

						// Process the optional Trailer field of a packet.
						case 8: {if (PacketInfo) fprintf(fp, "Trailer:  x%08lX\n", field); packet_counter++; current_field = 0; break;}
					}
				}
		}
		fclose(fp);
		return 0;
	}
	else
		return 1;
}

// Report the number of bytes currently stored in the DMA FIFO of the requested channel.
// Returns the number of samples.
unsigned long FIFOResidue(s_Adapter *p_HW, unsigned long ChannelNumber)
{
	unsigned long long Value = 0;
	
	Read(p_HW, BAR2, ChannelNumber << 12 | DATAPATH_MONITOR, &Value);
	
	return((unsigned long)((Value & 0x000000000000FFFFull)+((Value & 0x00000000FFFF0000ull)>>16)));
}

// Report the response time for the last interrupt that was serviced.
// Returns the response time in increments of 4 ns to a maximum value of (2^28)-1, or 1.074 sec.
// Note: The interrupt timeout (see below) will be reported if shorter than the actual response time.
unsigned long InterruptResponse(s_Adapter *p_HW)
{
	unsigned long long Value = 0;

	Read(p_HW, BAR0, INTERRUPT_LATENCY, &Value);

	return((unsigned long)Value);
}

// Set the interrupt timeout in increments of 4 ns to a maximum value of (2^28)-1, or 1.074 sec.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long InterruptTimeout(s_Adapter *p_HW, unsigned long IntTimeout)
{
	return(Write(p_HW, BAR0, INTERRUPT_TIMEOUT, (unsigned long long)IntTimeout));
}

// Benchmark the maximum attainable PCI performance of RX and/or TX DMA transactions.
// WARNING: DMA buffers must be allocated prior to executing this function or the computer will likely crash!
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long PciBenchmark(s_Adapter *p_HW)
{
	unsigned long long Value = 0;
	unsigned long long int_mask = 0;
	unsigned long Status = 0;
	unsigned long busy = 0;

	// Record the current state of the global interrupt mask.
	Status += Read(p_HW, BAR0, GLOBAL_INT_MASK, &int_mask);
	// Disable interrupts during the benchmark.
	Status += Write(p_HW, BAR0, GLOBAL_INT_MASK, 0x0);

	// Initiate continuous DMA transfers on all channels.
	Status += Write(p_HW, BAR0, PCI_THROUGHPUT, 0xFFull);

	// Pause while DMA transactions are counted for a quarter second.
	MicrosecondTimer(300000);

	// Discontinue continuous DMA transfers on the channels that participated in the benchmark.
	Status += Write(p_HW, BAR0, PCI_THROUGHPUT, 0x0ull);

	// Monitor the status of all channels with a polling loop and terminate when all channels stop processing.
	do
	{
		// Check the current operating state of each channel that participated in the benchmark to determine if any are still active.
		Read(p_HW, BAR0, DMA_1TO2_STATUS, &Value);
		busy = (Value | (Value >> 32)) & 0x00005000ull;
		Read(p_HW, BAR0, DMA_3TO4_STATUS, &Value);
		busy |= (Value | (Value >> 32)) & 0x00005000ull;
		Read(p_HW, BAR0, DMA_5TO6_STATUS, &Value);
		busy |= (Value | (Value >> 32)) & 0x00005000ull;
		Read(p_HW, BAR0, DMA_7TO8_STATUS, &Value);
		busy |= (Value | (Value >> 32)) & 0x00005000ull;
	}
	while (busy);

	// Restore the original interrupt mask state.
	Status += Write(p_HW, BAR0, GLOBAL_INT_MASK, int_mask);

	return(Status);
}

// Report the measured instantaneous PCI throughput for the last 250 ms and the maximum DMA latency since the last clear.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long PciPerformance(s_Adapter *p_HW, unsigned long *p_Performance)
{
	unsigned long long Value = 0;
	unsigned long Status = 0;

	// Read the register that stores PCI throughput every 250 ms.
	Status += Read(p_HW, BAR0, PCI_THROUGHPUT, &Value);

	// Record the measured receiver throughput in bytes/sec.
	p_Performance[0] = Value & 0x00000000FFFFFFFFull;
	// Record the measured transmitter throughput in bytes/sec.
	p_Performance[1] = (Value >> 32) & 0x00000000FFFFFFFFull;

	// Read the register that stores maximum PCI latency.
	Status += Read(p_HW, BAR0, PCI_LATENCY, &Value);

	// Record the measured receiver latency as the number of 4 ns clock periods.
	p_Performance[2] = Value & 0x00000000FFFFFFFFull;
	// Record the measured transmitter latency as the number of 4 ns clock periods.
	p_Performance[3] = (Value >> 32) & 0x00000000FFFFFFFFull;

	return(Status);
}

// Dump the contents of the transmitter datapath capture RAM.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ProbeSave(s_Adapter *p_HW, unsigned long ChannelNumber, char *p_File)
{
	FILE *fp;
	unsigned long long Value = 0;
	size_t record_count;
	
	// Open the specified data file.
	fp = fopen(p_File, "w");

	// Process the monitor contents only if the file open was successful.
	if(fp != NULL)
	{
		// Sequence through every 32-bit record in the 1024 deep memory.
		for(record_count = 0; record_count < 1024; record_count++)
		{
			// Write the memory address to access next.
			Write(p_HW, BAR2, ChannelNumber << 12 | DATAPATH_MONITOR, (unsigned long long)record_count);

			// Read the value stored at the memory address.
			Read(p_HW, BAR2, ChannelNumber << 12 | DATAPATH_MONITOR, &Value);

			// Sign extend negative numbers.
			if (Value & 0x0000000080000000ull) Value |= 0xFFFFFFFF00000000ull;
			//Print the value to the file.
			fprintf(fp, "%d\n", (signed)Value);
		}
		fclose(fp);
		return 0;
	}
	else
		return 1;
}

// Query each channel for any status change and update the user defined structure of the hardware handle accordingly.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ProcessChannels(s_Adapter *p_HW)
{
	unsigned long Status = 0;
	unsigned long CurrentMarker = 0;	// Variable to store the most marker number of the current DMA chapter.
	unsigned long chancnt = 0;			// Variable to index channel numbers in loops.
	unsigned long pagecnt = 0;			// Counter to sequence through the DMA pages of a chapter.
	s_User *User = NULL;				// Pointer to the user defined structure of the hardware handle.
	s_ChannelStatus ChannelStatus;		// Structure to store the results returned by a channel status query to the device.
	s_DMAStatus DMAStatus;				// Structure to store the results returned by a DMA status query to the device.
	s_GlobalStatus GlobalStatus;		// Structure to store the results returned by an interrupt status query to the device.

	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// Determine which channels have a status update.
	Status = GetGlobalStatus(p_HW, &GlobalStatus);

	// Update the user defined structure of any channel reporting a status change.
	for (chancnt = 1; chancnt <= User->RXChannelCount + User->TXChannelCount; chancnt++)
	{
		if (GlobalStatus.Channel[chancnt-1])
		{
			// Determine which channel status events require an update.
			Status += GetChannelStatus(p_HW, chancnt, &ChannelStatus);

			// Increment the converter (ADC/DAC) error count if an error is detected.
			if (ChannelStatus.Error[0]) User->Error[chancnt-1][0]++;

			// Increment the FIFO error count if an error is detected.
			if (ChannelStatus.Error[1]) User->Error[chancnt-1][1]++;

			// Process the DMA buffer contents if the DMA refresh marker is detected.
			if (ChannelStatus.DMAMarker)
			{
				// Copy the contents of the last chapter of DMA pages that were refreshed between DMA markers into data buffers accessible to the main application.
				// Note: This assumes that there was only one DMA marker update since the last check.
				for (pagecnt = 0; pagecnt < p_HW->DMA.PagesPerMark[chancnt-1]; pagecnt++)
					// Determine whether the current channel is a transmitter or receiver.
					if (chancnt > User->RXChannelCount)
						// Copy from the data buffer to the DMA buffer if this is a transmitter channel.
						memcpy(p_HW->DMA.VirtDMAAdr[chancnt-1][(User->MarkerIndex[chancnt-1]*p_HW->DMA.PagesPerMark[chancnt-1])+pagecnt], User->p_DataBuffer[chancnt-1][pagecnt], p_HW->DMA.BurstSizeB[chancnt-1]*p_HW->DMA.BurstCount[chancnt-1]);
					else
						// Copy from the DMA buffer to the data buffer if this is a receiver channel.
						memcpy(User->p_DataBuffer[chancnt-1][pagecnt], p_HW->DMA.VirtDMAAdr[chancnt-1][(User->MarkerIndex[chancnt-1]*p_HW->DMA.PagesPerMark[chancnt-1])+pagecnt], p_HW->DMA.BurstSizeB[chancnt-1]*p_HW->DMA.BurstCount[chancnt-1]);

				// Determine which channel status events require an update.
				Status += GetDMAStatus(p_HW, chancnt, &DMAStatus);
				// Compute the current marker index based on the current page of the DMA engine.
				CurrentMarker = (unsigned long)floor(DMAStatus.CurrentPage/p_HW->DMA.PagesPerMark[chancnt-1]);

				// Determine whether the current marker is beyond the next expected marker, which means that data is already lost.
				// There is a roll over condition to consider when the last DMA page is completed and the index starts back at zero.
				// Note: This technique cannot detect loss of data if the DMA engine completes a full cycle through all of the available pages before the test is performed.
				if ((CurrentMarker > User->MarkerIndex[chancnt-1] + 1) || (DMAStatus.Engine[0] && CurrentMarker == 0 && User->MarkerIndex[chancnt-1] > 0 && User->MarkerIndex[chancnt-1] < User->MaxMarkerIndex[chancnt-1]))
					// Increment the DMA error count if an error is detected.
					User->Error[chancnt-1][2]++;

				// Set the marker index to the current marker position.
				User->MarkerIndex[chancnt-1] = CurrentMarker;
				
			}
		}
	}

	return(Status);
}

// Initialize the structure members assigned to the user defined pointer attached to the hardware handle.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long UserInitialize(s_User *p_User, unsigned long ChannelNumber, s_DMABuffer DMABuffer, unsigned long *p_DataBuffer)
{
	unsigned long Status = 0;
	size_t index = 0;
	
	// Initialize the DMA marker index to the first set of pages.
	p_User->MarkerIndex[ChannelNumber-1] = 0;
	// Compute the maximum marker index based on the DMA buffer settings.
	p_User->MaxMarkerIndex[ChannelNumber-1] = (DMABuffer.PageCount / DMABuffer.PagesPerMark) - 1;
	for (index = 0; index < DMABuffer.PagesPerMark; index++)
	{
		p_User->p_DataBuffer[ChannelNumber-1][index] = malloc(DMABuffer.BurstSizeB * DMABuffer.BurstCount);
		if (p_User->p_DataBuffer[ChannelNumber-1][index] == NULL)
		{
			// Do not continue because there is no place to store data.
			Status++;
			// Mark this data buffer as a failed allocation.
			p_DataBuffer[index] = 1;
		}
		else
			// Mark this data buffer as a successful allocation.
			p_DataBuffer[index] = 0;
	}
	// Reset the error counters.
	for (index = 0; index < 3; index++)
		p_User->Error[ChannelNumber-1][index] = 0;

	return(Status);
}
